/*

	Import data from a Comma Separated Value (CSV) file
	
*/

#include "Hackdiet.h"

#ifdef CSV

/*	Defining CSV_SLOPPY allows importing CSV files which contain omitted
	trailing fields.  From the term SLOPPY, you can guess where the need
	for this originated--recent versions of Excel shipped with Microsoft
	Orifice 97 and 2000 which "optimise" CSV output in this manner, no
	doubt torpedoing thousands of third-party applications, which is
	almost certainly the goal.
	
	We always permit omitted trailing fields in new-format CSV files
	since we want to encourage people to process them in various ways
	and don't want to enforce a rigid format on them.  */
	
#define CSV_SLOPPY

#ifdef CSVDEBUG
int lastCSVfield = 0;
#endif

static DmOpenRef idb = NULL;
static MemHandle recHandle = NULL;
static char ImportDbName[34];					// Current database name
static UInt16 cardno;							// Current database card number
static LocalID dbid;							// Current database local ID

//	parseDayNew  --  Parse a record from an new-format (HDread) CSV database

static Boolean parseDayNew(char *s,				// 			weight as a string / scan buffer
						   weight *wgt,			//			weight
						   rung *rng,			//			rung
						   Boolean *flag,		//			flag
						   char *comment)		//			comment
{
	char *d1;
	int n1, n2;
	
	/*	We arrive here from parseDay with the date already stored
		into the result variables, the weight in the string s, and
		the CSV parser ready to scan the third (rung) field.  */

	//	Field 2: weight  (already scanned into s)
	
	if (StrLen(s) > 0) {
		if (isDigit(s[0])) {
			d1 = StrChr(s, '.');
			n2 = 0;
			if (d1 != NULL) {
				*d1++ = 0;
				//	Truncate or pad as required to obtain two decimal places
				if (StrLen(d1) > 2) {
					d1[2] = 0;
				}
				while (StrLen(d1) < 2) {
					StrCat(d1, "0");
				}
				n2 = StrAToI(d1);
			}
			n1 = StrAToI(s);
			*wgt = (n1 * 100) + n2;
		}
	}
	
	//	Field 3: rung
	
	if (!CSVscanField(s)) {
		return true;
	}
	if ((StrLen(s) > 0) && isDigit(s[0])) {
		*rng = StrAToI(s);
		//	Validate specified rung is in range
		if (*rng < 1 || *rng > RUNG_MAX) {
			return false;
		}
	} else {
		//	Mark rung absent
		*rng = RUNG_ABSENT;
	}
#ifdef CSVDEBUG
	lastCSVfield++;
#endif

	//	Field 4: flag
	
	if (!CSVscanField(s)) {
		return true;
	}
#ifdef CSVDEBUG
	lastCSVfield++;
#endif
	if (StrLen(s) > 0) {
		*flag = s[0] != '0';
	}
	
	//	Field 5: comment
	
	if (CSVscanField(s)) {
		s[MAX_COMMENT] = 0;					// Truncate comment if too long in CSV
		StrCopy(comment, s);
#ifdef CSVDEBUG
		lastCSVfield++;
#endif
	}
	
	return true;
}

//	parseDay  --  Parse a record from an old-format (Excel) CSV database

Boolean parseDay(char *csvLine,						// Input
				 int *year, int *month, int *day,	// Outputs: date
				 weight *wgt,						//			weight
				 rung *rng,							//			rung
				 Boolean *flag,						//			flag
				 char *comment)						//			comment
{
	char s[256];
	char *d1, *d2;
	char delim;
	int n1, n2, n3;
	
	*year = -1;
	*month = -1;
	*day = -1;
	*wgt = 0;
	*rng = RUNG_ABSENT;
	*flag = false;
	comment[0] = 0;
	
#ifdef CSVDEBUG
	lastCSVfield = 0;
#endif
	
	CSVscanInit(csvLine);
	
	//	Parse date.  Non-data lines are discarded at this phase
	
	if (!CSVscanField(s)) {
		return false;
	}
#ifdef CSVDEBUG
	lastCSVfield++;
#endif

	
	/*	We accept dates in any of the following forms:
	
			U.S.		m/d/yy	m/d/yyyy
			European	d.m.yy	d.m.yyyy
			ISO 8601	yyyy-mm-dd
			ISO Idiots	yy-mm-dd
			
		Two digit dates are converted using the convention that
		70-99 map to 1970-1999 and 0 to 69 to 2000-2069.
			
	*/
	
	d1 = StrChr(s, '.');
	if (d1 == NULL) {
		d1 = StrChr(s, '/');
	}
	if (d1 == NULL) {
		d1 = StrChr(s, '-');
	}
	if (d1 == NULL) {
		return false;			//	No date delimiter found
	}
	d2 = StrChr(d1 + 1, *d1);
	if (d2 == NULL) {
		return false;			//	Second data delimiter missing
	}
	delim = *d1;
	*d1 = *d2 = 0;
	n1 = StrAToI(s);
	n2 = StrAToI(d1 + 1);
	n3 = StrAToI(d2 + 1);
	switch (delim) {
		case '-':				// ISO
			*year = n1;
			*month = n2;
			*day = n3;
			break;
			
		case '.':				// European
			*day = n1;
			*month = n2;
			*year = n3;
			break;
			
		case '/':				// U.S.
			*month = n1;
			*day = n2;
			*year = n3;
			break;
	}
	if (*year < 100) {
		*year += (*year < 70) ? 2000 : 1900;
	}
	
	//	Do some rudimentary validation of the date.
	
	if ((*day < 1) || (*day > DaysInMonth(*month, *year)) ||
		(*month < 1) || (*month > 12)) {
		return false;
	}
	
	/*	Field 2 is just the weekday, which we ignore.
		If it's absent, however, the record is malformed.
		This is the point at which we make the distinction
		between old format (Excel) and new format (HDread)
		CSV files.  The second field in a new format log
		is the weight, and hence will always be numeric or
		blank.  In an old format log it's the weekday, which
		always begins with a letter.  */
		
	if (!CSVscanField(s)) {
		return false;
	}
#ifdef CSVDEBUG
	lastCSVfield++;
#endif

	if ((s[0] == 0) || (!isAlpha(s[0]))) {
		return parseDayNew(s, wgt, rng, flag, comment)	;
	}

	//	Field 3: weight
	
	if (!CSVscanField(s)) {
		return false;
	}
#ifdef CSVDEBUG
	lastCSVfield++;
#endif
	if (StrLen(s) > 0) {
		if (isDigit(s[0])) {
			d1 = StrChr(s, '.');
			n2 = 0;
			if (d1 != NULL) {
				*d1++ = 0;
				//	Truncate or pad as required to obtain two decimal places
				if (StrLen(d1) > 2) {
					d1[2] = 0;
				}
				while (StrLen(d1) < 2) {
					StrCat(d1, "0");
				}
				n2 = StrAToI(d1);
			}
			n1 = StrAToI(s);
			*wgt = (n1 * 100) + n2;
		} else {
		
			/*	The convention in Excel logs was that comments
				appeared in the Weight field, to explain missing
				weights (for example, when out of town).  Thus we
				treat any entry which doesn't begin with a digit
				as a comment.  In new style logs the weight and
				comment fields are independent and both can
				appear for a given day.  */
			
			s[MAX_COMMENT] = 0;					// Truncate comment if too long in CSV
			StrCopy(comment, s);
		}
	}
	
	//	Field 4: trend  This is ignored because we recompute locally
	
	if (!CSVscanField(s)) {
		return false;
	}
#ifdef CSVDEBUG
	lastCSVfield++;
#endif
	
	//	Field 5: variance  This is ignored because we recompute locally
	
	if (!CSVscanField(s)) {
		return false;
	}
#ifdef CSVDEBUG
	lastCSVfield++;
#endif
	
	//	Field 6: filler
	
	if (!CSVscanField(s)) {
		return false;
	}
#ifdef CSVDEBUG
	lastCSVfield++;
#endif
	
	//	Field 7: rung
	
	if (!CSVscanField(s)) {
#ifdef CSV_SLOPPY
		return true;
#else
		return false;
#endif
	}
	if ((StrLen(s) > 0) && isDigit(s[0])) {
		*rng = StrAToI(s);
		//	Validate specified rung is in range
		if (*rng < 1 || *rng > RUNG_MAX) {
			return false;
		}
	} else {
		//	Mark rung absent
		*rng = RUNG_ABSENT;
	}
#ifdef CSVDEBUG
	lastCSVfield++;
#endif

	//	Field 8: flag
	
	if (!CSVscanField(s)) {
#ifdef CSV_SLOPPY
		return true;
#else
		return false;
#endif
	}
#ifdef CSVDEBUG
	lastCSVfield++;
#endif
	if (StrLen(s) > 0) {
		*flag = s[0] != '0';
	}
	return true;
}

//	importCSVstart  --  Initialise import of a sequence of lines from a CSV file

static int csvLineNo, cYear, cMonth;
static UInt16 cRecno;
static Boolean mDirty = false;
static struct hdMonth *m = NULL;

void importCSVstart(void)
{
	csvLineNo = 0;
	mDirty = false;
	cYear = cMonth = -1;
	m = NULL;
}

//	changeMonth  --  Prepare to change to a new month

static void changeMonth(void)
{
	if (m != NULL) {
		if (mDirty) {
			hdPutMonthUnpacked(m, cRecno);
			hdPropagateTrend(m->year, m->month);
			mDirty = false;
		}
		hdDisposeMonthUnpacked(m);
		m = NULL;
	}
}

//	needMonth  --  Bring month into memory when needed

static void needMonth(int year, int month)
{
	if ((year != cYear) || (month != cMonth) || (m == NULL)) {
		changeMonth();
		cYear = year;
		cMonth = month;
		m = hdGetMonthByDate(year, month - 1, &cRecno, true);
		mDirty = false;
	}
}

/*	importCSVline  --  Process one line of a CSV file.  Returns
							true	Line imported OK
							false	Line isn't a log entry
*/

Boolean importCSVline(char *l)
{
	Boolean ok;
	
	int year, month, day;
	weight wgt;
	rung rng;
	Boolean flag;
	char cmt[132];
	
	if ((ok = parseDay(l, &year, &month, &day, &wgt, &rng, &flag, cmt)) == true) {
		
		/*	Fill in data from CSV file.  We will not, however, overwrite
			nonblank fields in the database.  Because there's no present or
			absent value for the flags field (since they're stored as a
			bit vector in the packed database), we can't quite provide
			this protection for flags.  We will not, however, permit CSV
			import to unset a flag set on the handheld.  If a flag present
			bit vector is added to the packed month (which would only add
			4 bytes per month), this can be revisited.
			
			An additional twist is that we defer creating a database record
			for a month until the first time we need to store a non-default
			record into it.  This permits us to import CSV files containing
			blank lines for future dates without cluttering the database
			with blank month records for them.
			
		*/
		
		day--;
#define Assign(dest, src, blank)	if (((src) != (blank))) { \
										needMonth(year, month);	\
										if (((dest) == (blank)) && ((dest) != (src))) { \
											dest = src;	\
											mDirty = true;	\
										}	\
									}	

		Assign(m->weightArray[day], wgt, 0);
		Assign(m->rungArray[day], rng, RUNG_ABSENT);
		Assign(m->flagsArray[day], flag, 0);
		
#undef Assign
		if ((StrLen(cmt) > 0) && (m->commentArray[day] == 0)) {
			struct hdMonth *nm;
			
			needMonth(year, month);
			nm = hdAddComment(m, day, cmt);
			m = nm;
			mDirty = true;
		}
	}
	return ok;
}

//	importCSVend  --  Clean up after importing CSV file

void importCSVend(void)
{
	changeMonth();
}

/*	csvImportNextDatabase  --  Find the next database to import and open
							   it for reading.  If no databases remain to
							   be processed, idb will be NULL.  */
							   
static Boolean onerec, eof, canceled;
static UInt16 rindex, nrecs;
static char *dbdata;
							   
static void csvImportNextDatabase(void)
{
	static DmSearchStateType stateInfo;
	Err dbeof;
	FormPtr f = FrmGetActiveForm();
	char fmt[80]; // , s[80];
	static char importingString[64], numLinesString[48];
	
	//	If a database is open, close it, unlocking the current record, and delete it.

	if (idb != NULL) {
		if (recHandle != NULL) {
			MemHandleUnlock(recHandle);
			recHandle = NULL;
		}
		DmCloseDatabase(idb);
		idb = NULL;
		if (!canceled) {
			
			//	Import successfully completed--delete the import database
		
			DmDeleteDatabase(cardno, dbid);
		}
	}

	/*	Search for next database of this type.  Always calling
		DmGetNextDatabaseByTypeCreator with a newSearch argument
		of true seems curious until you recall that we delete each
		CSV database after importing it.  Hence, each call finds
		the first *remaining* database with this creator and
		type.  Deleting a database invalidates the stateInfo for
		the search, necessitating a new search every time.  */
	
	dbeof = DmGetNextDatabaseByTypeCreator(true,
				&stateInfo, ImportDbType, CreatorID, false,
				&cardno, &dbid);
	if (dbeof == 0) {
		DmDatabaseInfo(cardno, dbid, ImportDbName, NULL, NULL,
			NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL);
		ImportDbName[33] = 0;
	} else {
		return;
	}
	idb = DmOpenDatabase(cardno, dbid, dmModeReadOnly);
	if (idb != NULL) {
//		char *nl;
		
		rindex = 0;
		eof = canceled = false;
		dbdata = NULL;
		importCSVstart();
		nrecs = DmNumRecordsInCategory(idb, dmAllCategories);
					
		/*	If the database contains a single record, we take it to be
			a complete CSV file embedded as a single binary chunk with
			a 4 byte length at the start of the chunk.  Otherwise, we
			assume this to be a text mode embedded database with one
			record per line of CSV file.  This complication is
			unnecessary, as either mode would be sufficient, but it
			allows testing both embedding modes of the desktop
			"pdbmake" utility.  */
		
		if ((onerec = (nrecs == 1)) == true) {
			recHandle = DmQueryNextInCategory(idb, &rindex, dmAllCategories);
			if (recHandle != NULL) {
				dbdata = MemHandleLock(recHandle);
				csvRecordInit(dbdata + sizeof(long), *((long *) dbdata));
				nrecs = csvRecordCount();
			}
		}
		
		stringResource(CsvImportFileNameString, fmt, sizeof fmt);
		assert((StrLen(fmt) + 34) < (sizeof importingString));
		StrPrintF(importingString, fmt, ImportDbName);
#ifdef OLDWAY
		CtlSetLabel(FrmGetObjectPtr(f, FrmGetObjectIndex(f, CSVImportFileNameLabel)),
			importingString);
#else
		FrmCopyLabel(f, CSVImportFileNameLabel, importingString);
#endif

		stringResource(CsvNumLinesString, fmt, sizeof fmt);
		assert((StrLen(fmt) + 8) < (sizeof importingString));
		StrPrintF(numLinesString, fmt, nrecs);
#ifdef OLDWAY
		CtlSetLabel(FrmGetObjectPtr(f, FrmGetObjectIndex(f, CSVImportFileLinesLabel)),
			numLinesString);
		CtlSetLabel(FrmGetObjectPtr(f, FrmGetObjectIndex(f, CSVImportCurrentLineLabel)), "");
#else
		FrmCopyLabel(f, CSVImportFileLinesLabel, numLinesString);
		FrmCopyLabel(f, CSVImportCurrentLineLabel, "");
#endif
		FrmEraseForm(f);
		FrmDrawForm(f);
	}
}

/*	csvImportHandleEvent  --  Handle event while the CSV import form
							   is displayed.  */
							  

static Boolean csvImportHandleEvent(EventPtr event)
{

#define nextLineEvent	((eventsEnum) (firstUserEvent + 1))

	Boolean handled;
	FormPtr f;
	static ControlPtr ctl;
	RectangleType r;
	static char fmt[80];
	static char importCompleteString[36], doneString[12];
	char s[80];
#ifdef __GNUC__
	CALLBACK_PROLOGUE
#endif
	handled = false;
  	f = FrmGetActiveForm();

	switch (event->eType) {
	  	
  		case frmOpenEvent:
  			csvImportNextDatabase();
  			if (idb != NULL) {
				
				//	Iterate over lines in the file and import valid ones
								
				stringResource(CsvCurrentLineString, fmt, sizeof fmt);
				ctl = FrmGetObjectPtr(f, FrmGetObjectIndex(f, CSVImportCurrentLineLabel));
				
				/*	Queue the first nextLineEvent to start the cascade of
					self-triggering events to import successive lines from
					the CSV database.  */

				event->eType = nextLineEvent;
				EvtAddEventToQueue(event);
  			}
			handled = true;
			break;
			
		case nextLineEvent:
			{
				char *dbdata = NULL;
				Boolean importOK;
				char csvRecord[132];
				
				while (!eof) {
					csvRecord[0] = 0;
					if (onerec) {
						eof = !csvRecordNext(csvRecord, sizeof csvRecord);
					} else {
						if (rindex < nrecs) {
							recHandle = DmQueryRecord(idb, rindex);
							if (recHandle != NULL) {
								dbdata = MemHandleLock(recHandle);
								StrCopy(csvRecord, dbdata);
								MemHandleUnlock(recHandle);
								recHandle = NULL;
								eof = false;
							} else {
								eof = true;
							}
						} else {
							eof = true;
						}
					}
					if (!eof) {
						importOK = importCSVline(csvRecord);
						rindex++;
						StrPrintF(s, fmt, rindex);
						RctSetRectangle(&r, ctl->bounds.topLeft.x, ctl->bounds.topLeft.y, 80, 15);
						drawInField(s, &r);
						if (canceled) {
							eof = true;
							break;
						} else {
							event->eType = nextLineEvent;
							EvtAddEventToQueue(event);
							handled = true;
							goto epilogue;
						}						
					} else {
						importCSVend();
						csvImportNextDatabase();
						if (idb != NULL) {
							eof = false;
							event->eType = nextLineEvent;
							EvtAddEventToQueue(event);
							handled = true;
							goto epilogue;
						}
					}
				}

				stringResource(CsvImportCompleteString, importCompleteString, sizeof importCompleteString);
#ifdef OLDWAY
				CtlSetLabel(FrmGetObjectPtr(f, FrmGetObjectIndex(f, CSVImportCurrentLineLabel)),
					importCompleteString);
#else
				FrmCopyLabel(f, CSVImportCurrentLineLabel, importCompleteString);
#endif
				stringResource(CsvImportDoneString, doneString, sizeof doneString);
				CtlSetLabel(FrmGetObjectPtr(f, FrmGetObjectIndex(f, CSVImportCancelButton)),
					doneString);
				FrmDrawForm(f);
			}
			handled = true;
			break;
			
		case frmCloseEvent:
			FrmEraseForm(FrmGetActiveForm());
			importCSVend();
			if (idb != NULL) {
				if (recHandle != NULL) {
					MemHandleUnlock(recHandle);
					recHandle = NULL;
				}
				DmCloseDatabase(idb);
				idb = NULL;
				if (!canceled) {
					LocalID lid;
					
					//	Import successfully completed--delete the import database
				
					lid = DmFindDatabase(0, ImportDbName);
					DmDeleteDatabase(0, lid);
				}
			}
			FrmEraseForm(FrmGetActiveForm());
			FrmDeleteForm(FrmGetActiveForm());
			handled = true;
			break;
			
		case ctlSelectEvent:
			switch (event->data.ctlSelect.controlID) {
				case CSVImportCancelButton:			// Done button: Back to month list
					if (eof) {
						FrmGotoForm(YearListForm);
					} else {						// Cancel button: Set import canceled flag
						canceled = true;
					}
					handled = true;
					break;
										
			}
			break;
	}

epilogue:;
#ifdef __GNUC__
	CALLBACK_EPILOGUE
#endif
	return handled;
}

//	csvImportLoad  --  Initialise CSV import form

void csvImportLoad(FormPtr f)
{
	FrmSetEventHandler(f, csvImportHandleEvent);
}

/*	checkImportCSV  --  See if an import file is present and
						process it if so.  */
						
Boolean checkImportCSV(void)
{
	DmOpenRef dbp;
	
	dbp = DmOpenDatabaseByTypeCreator(ImportDbType, CreatorID, dmModeReadOnly);
	if (dbp != NULL) {
		DmCloseDatabase(dbp);
		return true;
	}
	return false;
}

#endif